# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0002_auto_20170125_1127'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='company',
            options={'verbose_name_plural': 'companies'},
        ),
        migrations.AlterModelOptions(
            name='ticketstatus',
            options={'verbose_name_plural': 'ticket statuses'},
        ),
        migrations.AlterField(
            model_name='member',
            name='identifier',
            field=models.CharField(max_length=15, unique=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0010_auto_20170220_0901'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ticketpriority',
            name='color',
            field=models.CharField(db_column='color', max_length=50, blank=True, null=True),
        ),
    ]



import logging

from dateutil.parser import parse

from djconnectwise.api import CompanyAPIClient
from djconnectwise.api import ServiceAPIClient
from djconnectwise.api import SystemAPIClient
from djconnectwise.models import Company
from djconnectwise.models import ConnectWiseBoard
from djconnectwise.models import ConnectWiseBoardStatus
from djconnectwise.models import Member
from djconnectwise.models import Project
from djconnectwise.models import ServiceTicket
from djconnectwise.models import ServiceTicketAssignment
from djconnectwise.models import SyncJob
from djconnectwise.models import TicketPriority
from djconnectwise.models import TicketStatus
from djconnectwise.utils import get_hash, get_filename_extension

from django.core.exceptions import ObjectDoesNotExist
from django.core.files.base import ContentFile
from django.utils import timezone

DEFAULT_AVATAR_EXTENSION = 'jpg'

logger = logging.getLogger(__name__)


class InvalidStatusError(Exception):
    pass


class CompanySynchronizer:
    """
    Coordinates retrieval and demarshalling of ConnectWise JSON
    Company instances.
    """

    def __init__(self, *args, **kwargs):
        self.client = CompanyAPIClient()
        self.companies = self.load_company_dict()

    def load_company_dict(self):
        """
        Returns a dict of all companies that reside locally
        """
        return {c.id: c for c in Company.objects.all()}

    def _assign_field_data(self, company, api_company):
        """
        Assigns field data from an api_company instance
        to a local Company model instance
        """
        company.company_name = api_company['name']
        company.company_identifier = api_company['identifier']
        company.phone_number = api_company['phoneNumber']
        company.fax_number = api_company['faxNumber']
        company.address_line1 = api_company['addressLine1']
        company.address_line2 = api_company['addressLine1']
        company.city = api_company['city']
        company.state_identifier = api_company['state']
        company.zip = api_company['zip']
        company.created = timezone.now()

    def sync(self):
        api_company_data = self.client.get()
        created_count = 0
        updated_count = 0
        deleted_count = 0

        for api_company in api_company_data:
            company_id = api_company['id']
            try:
                company = Company.objects.get(id=company_id)
                updated_count += 1
            except ObjectDoesNotExist:
                company = Company.objects.create(id=company_id)
                created_count += 1

            self._assign_field_data(company, api_company)
            company.save()

        return created_count, updated_count, deleted_count

    def get_or_create_company(self, company_id):
        """
        Creates and returns a Company instance if it does not already exist
        """
        # Assign company to ticket. Create a new company if it does not already
        # exist

        company = self.companies.get(company_id)
        if not company:
            # fetch company from api
            api_company = self.client.by_id(company_id)
            company = Company.objects.create(id=company_id)

            self._assign_field_data(company, api_company)
            company.save()
            logger.info('Company Created: {}'.format(company_id))
            self.companies[company.id] = company
        return company


class BoardSynchronizer:

    def __init__(self, *args, **kwargs):
        self.client = ServiceAPIClient()

    def sync(self):
        updated_count = 0
        created_count = 0
        deleted_count = 0

        for board in self.client.get_boards():
            _, created = ConnectWiseBoard.objects.update_or_create(
                board_id=board['id'],
                defaults={
                    'name': board['name'],
                    'inactive': board['inactive'],
                }
            )

            if created:
                created_count += 1
            else:
                updated_count += 1

        return created_count, updated_count, deleted_count


class BoardStatusSynchronizer:

    def __init__(self, *args, **kwargs):
        self.client = ServiceAPIClient()

    def _sync(self, board_ids):

        updated_count = 0
        created_count = 0
        deleted_count = 0

        for board_id in board_ids:
            # TODO - Django doesn't provide an efficient
            # way to bulk get or create. May need to
            # invest time in a more efficient approach
            for status in self.client.get_statuses(board_id):
                _, created = ConnectWiseBoardStatus.objects.update_or_create(
                    status_id=status['id'],
                    defaults={
                        'board_id': board_id,
                        'status_name': status['name'],
                    }
                )

                if created:
                    created_count += 1
                else:
                    updated_count += 1

        return created_count, updated_count, deleted_count

    def sync(self, board_ids=None):

        if not board_ids:
            board_qs = ConnectWiseBoard.objects.all()
            board_ids = board_qs.values_list('board_id', flat=True)

        return self._sync(board_ids)


class PrioritySynchronizer:

    def __init__(self, *args, **kwargs):
        self.client = ServiceAPIClient()
        self.ticket_priority_map = {
            ticket.name: ticket for ticket in TicketPriority.objects.all()
        }

    def _assign_field_data(self, ticket_priority, api_priority):
        ticket_priority.name = api_priority['name']
        ticket_priority.priority_id = api_priority['id']
        ticket_priority.color = api_priority.get('color')
        ticket_priority.sort = api_priority.get('sortOrder')
        return ticket_priority

    def get_or_create_ticket_priority(self, api_priority):
        """
        Gets or creates a TicketPriority instance for
        the supplied API Priority JSON structure
        """
        ticket_priority = self.ticket_priority_map.get(
            api_priority['name'])

        created = False
        if not ticket_priority:
            ticket_priority = TicketPriority()
            self._assign_field_data(ticket_priority, api_priority)
            ticket_priority.save()
            created = True

        return ticket_priority, created

    def update_or_create_ticket_priority(self, api_priority):
        """
        Creates and returns a TicketPriority instance if
        it does not already exist
        """
        ticket_priority, created = self.get_or_create_ticket_priority(
            api_priority)

        action = 'Created' if created else 'Updated'

        if not created:
            self._assign_field_data(ticket_priority, api_priority)
            ticket_priority.save()

        self.ticket_priority_map[ticket_priority.name] = ticket_priority

        msg = 'TicketPriority {}: {}'
        logger.info(msg.format(action, ticket_priority.name))

        return ticket_priority, created

    def sync(self):
        created_count = 0
        updated_count = 0
        deleted_count = 0

        for api_priority in self.client.get_priorities():
            _, created = self.update_or_create_ticket_priority(api_priority)

            if created:
                created_count += 1
            else:
                updated_count += 1

        return created_count, updated_count, deleted_count


class MemberSynchronizer:

    def __init__(self, *args, **kwargs):
        self.client = SystemAPIClient()
        self.last_sync_job = None

        sync_job_qset = SyncJob.objects.all()

        if sync_job_qset.exists():
            self.last_sync_job = sync_job_qset.last()

    def _save_avatar(self, member, avatar, attachment_filename):
        """
        The Django ImageField (and ThumbnailerImageField) field adjusts our
        filename if the file already exists- it adds some random characters at
        the end of the name. This means if we just save a new image when the
        old one still exists, we'll get a new image for each save, resulting
        in lots of unnecessary images. So we'll delete the old image first,
        and then the save will use the exact name we give it.

        Well, except in the case where two or more members share the same
        image, because we're using content hashes as names, and ConnectWise
        gives users a common default avatar. In that case, the first save
        will use the expected name, while subsequent saves for other members
        will have some random characters added to the filename.

        This method tells Django not to call save() on the given model,
        so the caller must be sure to do that itself.
        """
        extension = get_filename_extension(attachment_filename)
        filename = '{}.{}'.format(
            get_hash(avatar), extension or DEFAULT_AVATAR_EXTENSION)
        avatar_file = ContentFile(avatar)
        member.avatar.delete(save=False)
        member.avatar.save(filename, avatar_file, save=False)
        logger.info("Saved member '{}' avatar to {}.".format(
            member.identifier, member.avatar.name))

    def sync(self):
        members_json = self.client.get_members()

        updated_count = 0
        created_count = 0
        deleted_count = 0

        for api_member in members_json:
            username = api_member['identifier']
            member_qset = Member.objects.filter(identifier=username)
            if member_qset.exists():
                member = member_qset.first()

                member.first_name = api_member['firstName']
                member.last_name = api_member['lastName']
                member.office_email = api_member['officeEmail']
                updated_count += 1
                logger.info('Update Member: {0}'.format(member.identifier))
            else:
                member = Member.create_member(api_member)
                created_count += 1
                logger.info('Create Member: {0}'.format(member.identifier))

            # only update the avatar if the member profile
            # was updated since last sync
            member_last_updated = parse(api_member['_info']['lastUpdated'])
            member_stale = False

            if self.last_sync_job:
                member_stale = member_last_updated > \
                    self.last_sync_job.start_time

            if not self.last_sync_job or member_stale:
                (attachment_filename, avatar) = self.client \
                    .get_member_image_by_identifier(username)
                if attachment_filename and avatar:
                    self._save_avatar(member, avatar, attachment_filename)

            member.save()

        return created_count, updated_count, deleted_count


class ServiceTicketSynchronizer:
    """
    Coordinates retrieval and demarshalling of ConnectWise JSON
    objects to the local counterparts.
    """
    def __init__(self, reset=False):
        self.company_synchronizer = CompanySynchronizer()
        self.status_synchronizer = BoardStatusSynchronizer()
        self.priority_synchronizer = PrioritySynchronizer()

        self.reset = reset
        self.last_sync_job = None
        extra_conditions = ''
        sync_job_qset = SyncJob.objects.all()

        if sync_job_qset.exists() and not self.reset:
            self.last_sync_job = sync_job_qset.last()
            last_sync_job_time = self.last_sync_job.start_time.isoformat()
            extra_conditions = "lastUpdated > [{0}]".format(last_sync_job_time)

            log_msg = 'Preparing sync job for objects updated since {}.'
            logger.info(log_msg.format(last_sync_job_time))
            logger.info(
                'ServiceTicket Extra Conditions: {0}'.format(extra_conditions))
        else:
            logger.info('Preparing full ticket sync job.')
            # absence of a sync job indicates that this is an initial/full
            # sync, in which case we do not want to retrieve closed tickets
            extra_conditions = 'ClosedFlag = False'

        self.service_client = ServiceAPIClient(
            extra_conditions=extra_conditions)

        self.system_client = SystemAPIClient()

        # we need to remove the underscores to ensure an accurate
        # lookup of the normalized api fieldnames
        self.local_service_ticket_fields = self._create_field_lookup(
            ServiceTicket)
        self.local_company_fields = self._create_field_lookup(Company)

        self.ticket_status_map = {
            ticket.status_name: ticket for ticket in TicketStatus.objects.all()
        }

        self.members_map = {m.identifier: m for m in Member.objects.all()}
        self.project_map = {p.name: p for p in Project.objects.all()}
        self.ticket_assignments = {}
        self.updated_members = []

        self.exclude_fields = ('priority', 'status', 'company')

    def _create_field_lookup(self, clazz):
        field_map = [
            (f.name, f.name.replace('_', '')) for
            f in clazz._meta.get_fields(
                include_parents=False, include_hidden=True)
        ]
        return dict(field_map)

    def _manage_member_assignments(self, service_ticket):
        # reset board/ticket assignment in case the assigned resources have
        # changed since last sync
        member = None
        if service_ticket.resources:
            usernames = [u.strip()
                         for u in service_ticket.resources.split(',')]
            # clear existing assignments
            ServiceTicketAssignment.objects.filter(
                service_ticket=service_ticket).delete()
            for username in usernames:
                member = self.members_map.get(username)
                if member:
                    assignment = ServiceTicketAssignment()
                    assignment.member = member
                    assignment.service_ticket = service_ticket
                    self.ticket_assignments[
                        (username, service_ticket.id,)] = assignment
                    logger.info('Member ServiceTicket Assignment: %s - %s' %
                                (username, service_ticket.id))
                else:
                    logger.error(
                        'Failed to locate member with username {} for ticket '
                        '{} assignment.'.format(username, service_ticket.id)
                    )

    def get_or_create_project(self, api_ticket):
        api_project = api_ticket['project']
        if api_project:
            project = self.project_map.get(api_project['name'])
            if not project:
                project = Project()
                project.name = api_project['name']
                project.project_href = api_project['_info']['project_href']
                project.project_id = api_project['id']
                project.save()
                self.project_map[project.name] = project
                logger.info('Project Created: %s' % project.name)
            return project

    def get_or_create_ticket_status(self, api_ticket):
        """
        Creates and returns a TicketStatus instance if
        it does not already exist
        """
        api_status = api_ticket['status']
        status_name = api_status['name'].strip()
        ticket_status = self.ticket_status_map.get(status_name)

        if not ticket_status:
            kwargs = dict(
                status_id=api_status['id'],
                status_name=status_name
            )

            if TicketStatus.objects.filter(**kwargs).exists():
                ticket_status = TicketStatus.objects.get(**kwargs)
            else:
                logger.info('TicketStatus Created - %s' % status_name)
                ticket_status, created = TicketStatus.objects.get_or_create(
                    **kwargs)

            self.ticket_status_map[status_name] = ticket_status

        else:
            if ticket_status.status_id != api_status['id']:
                ticket_status.status_id = api_status['id']
                ticket_status.save()

        return ticket_status

    def sync_ticket(self, api_ticket):
        """
        Creates a new local instance of the supplied ConnectWise
        ServiceTicket instance.
        """
        api_ticket_id = api_ticket['id']
        service_ticket, created = ServiceTicket.objects \
                                               .get_or_create(pk=api_ticket_id)

        # if the status results in a move to a different column
        original_status = not created and service_ticket.status or None
        new_ticket_status = self.get_or_create_ticket_status(api_ticket)

        service_ticket.closed_flag = api_ticket['closedFlag']
        service_ticket.type = api_ticket['type']
        service_ticket.priority_text = api_ticket['priority']['name']
        service_ticket.location = api_ticket['serviceLocation']
        service_ticket.summary = api_ticket['summary']
        service_ticket.entered_date_utc = api_ticket['dateEntered']
        service_ticket.last_updated_utc = api_ticket['_info']['lastUpdated']
        service_ticket.resources = api_ticket['resources']
        service_ticket.budget_hours = api_ticket['budgetHours']
        service_ticket.actual_hours = api_ticket['actualHours']
        service_ticket.record_type = api_ticket['recordType']

        team = api_ticket['team']
        if team:
            service_ticket.team_id = api_ticket['team']['id']

        service_ticket.api_text = str(api_ticket)
        service_ticket.board_name = api_ticket['board']['name']
        service_ticket.board_id = api_ticket['board']['id']
        service_ticket.board_status_id = api_ticket['status']['id']

        company_id = api_ticket['company']['id']
        service_ticket.company = self.company_synchronizer \
                                     .get_or_create_company(company_id)

        priority, _ = self.priority_synchronizer \
            .get_or_create_ticket_priority(api_ticket['priority'])
        service_ticket.priority = priority

        service_ticket.status = new_ticket_status
        service_ticket.project = self.get_or_create_project(api_ticket)
        service_ticket.save()
        action = created and 'Created' or 'Updated'

        status_changed = ''
        if original_status != new_ticket_status:
            status_txt = 'Status Changed From: {} To: {}'
            status_changed = status_txt.format(original_status,
                                               new_ticket_status)

        log_info = '{} Ticket #: {} {}'
        logger.info(log_info.format(action, service_ticket.id, status_changed))

        self._manage_member_assignments(service_ticket)
        return service_ticket, created

    def sync(self):
        """
        Synchronizes tickets between the ConnectWise server and the
        local database. Synchronization is performed in batches
        specified in the DJCONNECTWISE_API_BATCH_LIMIT setting
        """
        sync_job = SyncJob.objects.create()

        created_count = 0
        updated_count = 0
        ticket_ids = []

        page = 0
        accumulated = 0

        logger.info('Synchronization Started')

        while True:
            logger.info('Processing Batch #: {}'.format(page))
            tickets = self.service_client.get_tickets(page=page)
            num_tickets = len(tickets)

            for ticket in tickets:
                ticket, created = self.sync_ticket(ticket)
                ticket_ids.append(ticket.id)

                if created:
                    created_count += 1
                else:
                    updated_count += 1

            page += 1
            accumulated += len(tickets)

            if not num_tickets:
                break

        logger.info('Saving Ticket Assignments')
        ServiceTicketAssignment.objects.bulk_create(
            list(self.ticket_assignments.values()))

        # now prune closed service tickets.
        logger.info('Deleting Closed Tickets')
        delete_qset = ServiceTicket.objects.filter(closed_flag=True)
        delete_count = delete_qset.count()
        delete_qset.delete()

        sync_job.end_time = timezone.now()
        sync_job.save()

        return created_count, updated_count, delete_count


class ServiceTicketUpdater(object):
    """Send ticket updates to ConnectWise."""

    def __init__(self):
        self.service_client = ServiceAPIClient()

    def update_api_ticket(self, service_ticket):
        api_service_ticket = self.service_client.get_ticket(service_ticket.id)

        if service_ticket.closed_flag:
            api_service_ticket['closedFlag'] = service_ticket.closed_flag
            ticket_status, created = TicketStatus.objects.get_or_create(
                status_name__iexact='Closed')
        else:
            ticket_status = service_ticket.status

        if not service_ticket.closed_flag:
            # Ensure that the new status is valid for the CW board.
            try:
                cw_board = ConnectWiseBoard.objects.get(
                    id=service_ticket.board_id)
                board_status = ConnectWiseBoardStatus.objects.get(
                    board_id=service_ticket.board_id,
                    status_name=ticket_status.status_name
                )
                api_service_ticket['status']['id'] = board_status.status_id
            except ConnectWiseBoard.DoesNotExist:
                raise InvalidStatusError("Failed to find the ticket's board.")
            except ConnectWiseBoardStatus.DoesNotExist:
                raise InvalidStatusError(
                    "{} is not a valid status for the ticket's "
                    "ConnectWise board ({}).".
                    format(
                        ticket_status.status_name,
                        cw_board.name
                    )
                )

        # No need for a callback update when updating via api
        api_service_ticket['skipCallback'] = True
        logger.info(
            'Update API Ticket Status: {} - {}'.format(
                service_ticket.id, api_service_ticket['status']['name']
            )
        )

        return self.service_client.update_ticket(api_service_ticket)

    def close_ticket(self, service_ticket):
        """
        Closes the specified service ticket returns True if the close
        operation was successful on the connectwise server.
        Note: It appears that the connectwise server does not return a
        permissions error if user does not have access to this operation.
        """
        service_ticket.closed_flag = True
        service_ticket.save()
        logger.info('Close API Ticket: %s' % service_ticket.id)
        api_ticket = self.update_api_ticket(service_ticket)
        ticket_is_closed = api_ticket['closedFlag']

        if not ticket_is_closed:
            service_ticket.closed_flag = ticket_is_closed
            service_ticket.save()

        return ticket_is_closed



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0003_auto_20170125_1613'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='ticketpriority',
            name='description',
        ),
        migrations.RenameField(
            model_name='ticketpriority',
            old_name='title',
            new_name='name',
        ),
        migrations.AlterField(
            model_name='ticketpriority',
            name='name',
            field=models.CharField(max_length=50),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0011_auto_20170220_1013'),
    ]

    operations = [
        migrations.RenameField(
            model_name='ticketpriority',
            old_name='color',
            new_name='_color',
        ),
    ]



from copy import deepcopy
from unittest import TestCase

from djconnectwise.models import Company
from djconnectwise.models import ConnectWiseBoard
from djconnectwise.models import ConnectWiseBoardStatus
from djconnectwise.models import TicketPriority
from djconnectwise.models import Member
from djconnectwise.models import ServiceTicket

from . import fixtures
from . import mocks
from .. import sync


class TestCompanySynchronizer(TestCase):

    def setUp(self):
        self.synchronizer = sync.CompanySynchronizer()

    def _assert_fields(self, company, api_company):
        assert company.company_name == api_company['name']
        assert company.company_identifier == api_company['identifier']
        assert company.phone_number == api_company['phoneNumber']
        assert company.fax_number == api_company['faxNumber']
        assert company.address_line1 == api_company['addressLine1']
        assert company.address_line2 == api_company['addressLine1']
        assert company.city == api_company['city']
        assert company.state_identifier == api_company['state']
        assert company.zip == api_company['zip']

    def test_sync(self):
        _, get_patch = mocks.company_api_get_call(fixtures.API_COMPANY_LIST)
        self.synchronizer.sync()
        company_dict = {c['id']: c for c in fixtures.API_COMPANY_LIST}

        for company in Company.objects.all():
            api_company = company_dict[company.id]
            self._assert_fields(company, api_company)

    def test_sync_update(self):
        identifier = 'Some New Company'
        api_company = deepcopy(fixtures.API_COMPANY)
        api_company['identifier'] = identifier
        api_company_list = [api_company]
        company_pre_update = Company.objects \
                                    .get(id=api_company['id'])
        _, get_patch = mocks.company_api_get_call(api_company_list)

        self.synchronizer.sync()
        company_post_update = Company.objects \
                                     .get(id=api_company['id'])

        self.assertNotEqual(company_pre_update.company_identifier,
                            identifier)
        self._assert_fields(company_post_update, api_company)


class TestPrioritySynchronizer(TestCase):

    def setUp(self):
        self.synchronizer = sync.PrioritySynchronizer()
        self.valid_prio_colors = \
            list(TicketPriority.DEFAULT_COLORS.values()) + \
            [TicketPriority.DEFAULT_COLOR]

    def _assert_fields(self, priority, api_priority):
        assert priority.name == api_priority['name']
        assert priority.priority_id == api_priority['id']
        if 'color' in api_priority.keys():
            assert priority.color == api_priority['color']
        else:
            assert priority.color in self.valid_prio_colors
        if 'sortOrder' in api_priority.keys():
            assert priority.sort == api_priority['sortOrder']
        else:
            assert priority.sort is None

    def test_sync(self):
        _, get_patch = mocks.service_api_get_priorities_call(
            fixtures.API_SERVICE_PRIORITY_LIST)

        self.synchronizer.sync()
        priorities = fixtures.API_SERVICE_PRIORITY_LIST
        priority_dict = {p['id']: p for p in priorities}

        for priority in TicketPriority.objects.all():
            api_priority = priority_dict[priority.priority_id]
            self._assert_fields(priority, api_priority)

    def test_sync_update(self):
        color = 'green'
        api_priority = deepcopy(fixtures.API_SERVICE_PRIORITY)
        api_priority['color'] = color
        api_priority_list = [api_priority]
        priority_pre_update = TicketPriority.objects \
            .get(priority_id=api_priority['id'])
        _, get_patch = mocks.service_api_get_priorities_call(api_priority_list)

        self.synchronizer.sync()
        priority_post_update = TicketPriority.objects \
            .get(priority_id=api_priority['id'])

        self.assertNotEqual(priority_pre_update.color,
                            color)
        self._assert_fields(priority_post_update, api_priority)


class TestBoardSynchronizer(TestCase):

    def setUp(self):
        self.synchronizer = sync.BoardSynchronizer()

    def _sync(self):
        ConnectWiseBoard.objects.all().delete()
        mocks.service_api_get_boards_call(fixtures.API_BOARD_LIST)
        return self.synchronizer.sync()

    def _local_board_set(self):
        board_qs = ConnectWiseBoard.objects.all()
        return set(board_qs.values_list('board_id', 'name'))

    def _api_board_set(self, board_data):
        return set([(s['id'], s['name']) for s in board_data])

    def test_sync(self):
        created_count, updated_count, _ = self._sync()
        local_boards = self._local_board_set()
        api_boards = self._api_board_set(fixtures.API_BOARD_LIST)

        self.assertEqual(local_boards, api_boards)
        self.assertEqual(updated_count, 0)
        self.assertEqual(created_count, len(fixtures.API_BOARD_LIST))

    def test_sync_update(self):
        self._sync()
        updated_boards = deepcopy(fixtures.API_BOARD_LIST)
        updated_boards[0]['name'] = 'New Board Name'
        mocks.service_api_get_boards_call(updated_boards)
        created_count, updated_count, _ = self.synchronizer.sync()

        local_boards = self._local_board_set()
        api_boards = self._api_board_set(updated_boards)

        self.assertEqual(local_boards, api_boards)
        self.assertEqual(updated_count, len(fixtures.API_BOARD_LIST))
        self.assertEqual(created_count, 0)


class TestBoardStatusSynchronizer(TestCase):

    def setUp(self):
        self.synchronizer = sync.BoardStatusSynchronizer()

    def _board_ids(self, board_status_list):
        return [b['id'] for b in board_status_list]

    def _sync(self):
        ConnectWiseBoardStatus.objects.all().delete()
        mocks.service_api_get_statuses_call(fixtures.API_BOARD_STATUS_LIST)
        board_ids = self._board_ids(fixtures.API_BOARD_LIST)
        return self.synchronizer.sync(board_ids)

    def _assert_sync(self, board_status_list):
        status_qs = ConnectWiseBoardStatus.objects.all()
        local_statuses = set(status_qs.values_list('status_id', 'status_name'))
        api_statuses = set([(s['id'], s['name'])
                            for s in board_status_list])
        num_local_statuses = len(local_statuses)

        self.assertEqual(num_local_statuses, len(api_statuses))
        self.assertEqual(local_statuses, api_statuses)

    def test_sync_updated(self):
        self._sync()
        updated_statuses = deepcopy(fixtures.API_BOARD_STATUS_LIST)
        updated_statuses[0]['name'] = 'New Status Name'
        mocks.service_api_get_statuses_call(updated_statuses)

        board_ids = self._board_ids(fixtures.API_BOARD_LIST)

        created_count, updated_count, _ = self.synchronizer.sync(board_ids)

        self.assertEqual(created_count, 0)
        self.assertEqual(updated_count, len(fixtures.API_BOARD_STATUS_LIST))
        self._assert_sync(updated_statuses)

    def test_sync(self):
        created_count, updated_count, _ = self._sync()
        self.assertEqual(created_count, len(fixtures.API_BOARD_STATUS_LIST))
        self.assertEqual(updated_count, 0)
        self._assert_sync(fixtures.API_BOARD_STATUS_LIST)


class TestServiceTicketSynchronizer(TestCase):

    def setUp(self):
        self.synchronizer = sync.ServiceTicketSynchronizer()

    def _get_local_and_api_ticket(self):
        api_ticket = self.synchronizer.service_client.get_tickets()[0]

        local_ticket, created = self.synchronizer.sync_ticket(api_ticket)
        return local_ticket, api_ticket

    def _sync(self):
        mocks.company_api_by_id_call(fixtures.API_COMPANY)
        mocks.service_api_tickets_call()

        return self.synchronizer.sync()

    def test_sync(self):
        created_count, _, _ = self._sync()
        self.assertEqual(created_count, 1)


class TestServiceTicketUpdater(TestCase):
    def setUp(self):
        self.updater = sync.ServiceTicketUpdater()
        self.synchronizer = sync.ServiceTicketSynchronizer()

    def _sync(self):
        mocks.company_api_by_id_call(fixtures.API_COMPANY)
        mocks.service_api_tickets_call()

        return self.synchronizer.sync()

    def test_update_api_ticket(self):
        ServiceTicket.objects.all().delete()
        self._sync()

        board_name = 'Some Board Name'
        api_service_ticket = deepcopy(fixtures.API_SERVICE_TICKET)
        api_service_ticket['board']['name'] = board_name

        mocks.service_api_update_ticket_call(api_service_ticket)
        mocks.service_api_get_ticket_call()

        local_ticket = ServiceTicket.objects.first()
        local_ticket.board_name = board_name
        local_ticket.closed_flag = True
        local_ticket.save()

        updated_api_ticket = self.updater.update_api_ticket(local_ticket)

        self.assertEqual(
            updated_api_ticket['board']['name'],
            local_ticket.board_name
        )


class TestMemberSynchronization(TestCase):

    def setUp(self):
        self.member_id = 'User1'
        self.synchronizer = sync.MemberSynchronizer()
        mocks.system_api_get_members_call([fixtures.API_MEMBER])
        mocks.system_api_get_member_image_by_identifier_call(
            (mocks.CW_MEMBER_IMAGE_FILENAME, mocks.get_member_avatar()))

    def _assert_member_fields(self, local_member, api_member):
        self.assertEqual(local_member.first_name, api_member['firstName'])
        self.assertEqual(local_member.last_name, api_member['lastName'])
        self.assertEqual(local_member.office_email, api_member['officeEmail'])

    def _clear_members(self):
        Member.objects.all().delete()

    def test_sync_member_update(self):
        self._clear_members()
        member = Member()
        member.identifier = self.member_id
        member.first_name = 'some stale first name'
        member.last_name = 'some stale last name'
        member.office_email = 'some@stale.com'
        member.save()

        self.synchronizer.sync()
        local_member = Member.objects.get(identifier=self.member_id)
        api_member = fixtures.API_MEMBER
        self._assert_member_fields(local_member, api_member)

    def test_sync_member_create(self):
        self._clear_members()
        self.synchronizer.sync()
        local_member = Member.objects.all().first()
        api_member = fixtures.API_MEMBER
        self._assert_member_fields(local_member, api_member)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django_extensions.db.fields
import easy_thumbnails.fields


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CallBackEntry',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        serialize=False, verbose_name='ID', primary_key=True)),
                ('callback_type', models.CharField(max_length=25)),
                ('url', models.CharField(max_length=255)),
                ('level', models.CharField(max_length=255)),
                ('object_id', models.IntegerField()),
                ('entry_id', models.IntegerField()),
                ('member_id', models.IntegerField()),
                ('enabled', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='Company',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        serialize=False, verbose_name='ID', primary_key=True)),
                ('created', django_extensions.db.fields.CreationDateTimeField(
                    auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(
                    auto_now=True, verbose_name='modified')),
                ('company_name', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('company_alias', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('company_identifier', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('phone_number', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('fax_number', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('address_line1', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('address_line2', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('city', models.CharField(null=True, blank=True, max_length=250)),
                ('state_identifier', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('zip', models.CharField(null=True, blank=True, max_length=250)),
                ('country', models.CharField(null=True, blank=True, max_length=250)),
                ('type', models.CharField(null=True, blank=True, max_length=250)),
                ('status', models.CharField(null=True, blank=True, max_length=250)),
                ('territory', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('website', models.CharField(null=True, blank=True, max_length=250)),
                ('market', models.CharField(null=True, blank=True, max_length=250)),
                ('defaultcontactid', models.IntegerField(null=True, blank=True)),
                ('defaultbillingcontactid',
                 models.IntegerField(null=True, blank=True)),
                ('updatedby', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('lastupdated', models.CharField(
                    null=True, blank=True, max_length=250)),
            ],
            options={
                'get_latest_by': 'modified',
                'ordering': ('-modified', '-created'),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ConnectWiseBoard',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        serialize=False, verbose_name='ID', primary_key=True)),
                ('created', django_extensions.db.fields.CreationDateTimeField(
                    auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(
                    auto_now=True, verbose_name='modified')),
                ('board_id', models.PositiveSmallIntegerField()),
                ('name', models.CharField(max_length=255)),
                ('inactive', models.BooleanField()),
            ],
            options={
                'ordering': ('name',),
            },
        ),
        migrations.CreateModel(
            name='ConnectWiseBoardStatus',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        serialize=False, verbose_name='ID', primary_key=True)),
                ('created', django_extensions.db.fields.CreationDateTimeField(
                    auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(
                    auto_now=True, verbose_name='modified')),
                ('board_id', models.PositiveSmallIntegerField()),
                ('status_id', models.PositiveSmallIntegerField()),
                ('status_name', models.CharField(
                    null=True, blank=True, max_length=250)),
            ],
            options={
                'ordering': ('status_name',),
            },
        ),
        migrations.CreateModel(
            name='Member',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        serialize=False, verbose_name='ID', primary_key=True)),
                ('created', django_extensions.db.fields.CreationDateTimeField(
                    auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(
                    auto_now=True, verbose_name='modified')),
                ('first_name', models.CharField(max_length=30)),
                ('last_name', models.CharField(max_length=30)),
                ('identifier', models.CharField(max_length=15)),
                ('office_email', models.EmailField(max_length=250)),
                ('inactive', models.BooleanField(default=False)),
                ('avatar', easy_thumbnails.fields.ThumbnailerImageField(null=True, upload_to='',
                                                                        blank=True, verbose_name='Member Avatar', help_text='Member Avatar')),
            ],
            options={
                'get_latest_by': 'modified',
                'ordering': ('-modified', '-created'),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        serialize=False, verbose_name='ID', primary_key=True)),
                ('created', django_extensions.db.fields.CreationDateTimeField(
                    auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(
                    auto_now=True, verbose_name='modified')),
                ('name', models.CharField(max_length=200)),
                ('project_id', models.PositiveSmallIntegerField()),
                ('project_href', models.CharField(max_length=200)),
            ],
            options={
                'get_latest_by': 'modified',
                'ordering': ('-modified', '-created'),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ServiceProvider',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        serialize=False, verbose_name='ID', primary_key=True)),
                ('created', django_extensions.db.fields.CreationDateTimeField(
                    auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(
                    auto_now=True, verbose_name='modified')),
                ('title', models.CharField(verbose_name='title', max_length=255)),
                ('description', models.TextField(null=True,
                                                 blank=True, verbose_name='description')),
                ('slug', django_extensions.db.fields.AutoSlugField(blank=True,
                                                                   editable=False, verbose_name='slug', populate_from='title')),
                ('logo', easy_thumbnails.fields.ThumbnailerImageField(null=True, upload_to='',
                                                                      blank=True, verbose_name='Service Provider Logo', help_text='Service Provider Logo')),
            ],
            options={
                'get_latest_by': 'modified',
                'ordering': ('-modified', '-created'),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ServiceTicket',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        serialize=False, verbose_name='ID', primary_key=True)),
                ('created', django_extensions.db.fields.CreationDateTimeField(
                    auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(
                    auto_now=True, verbose_name='modified')),
                ('closed_flag', models.NullBooleanField()),
                ('type', models.CharField(null=True, blank=True, max_length=250)),
                ('sub_type', models.CharField(null=True, blank=True, max_length=250)),
                ('sub_type_item', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('priority_text', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('location', models.CharField(null=True, blank=True, max_length=250)),
                ('source', models.CharField(null=True, blank=True, max_length=250)),
                ('summary', models.CharField(null=True, blank=True, max_length=250)),
                ('entered_date_utc', models.DateTimeField(null=True, blank=True)),
                ('last_updated_utc', models.DateTimeField(null=True, blank=True)),
                ('resources', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('required_date_utc', models.DateTimeField(null=True, blank=True)),
                ('closed_date_utc', models.DateTimeField(null=True, blank=True)),
                ('site_name', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('budget_hours', models.DecimalField(null=True,
                                                     blank=True, decimal_places=2, max_digits=6)),
                ('actual_hours', models.DecimalField(null=True,
                                                     blank=True, decimal_places=2, max_digits=6)),
                ('approved', models.NullBooleanField()),
                ('closed_by', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('resolve_mins', models.IntegerField(null=True, blank=True)),
                ('res_plan_mins', models.IntegerField(null=True, blank=True)),
                ('respond_mins', models.IntegerField(null=True, blank=True)),
                ('updated_by', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('record_type', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('team_id', models.IntegerField(null=True, blank=True)),
                ('agreement_id', models.IntegerField(null=True, blank=True)),
                ('severity', models.CharField(null=True, blank=True, max_length=250)),
                ('impact', models.CharField(null=True, blank=True, max_length=250)),
                ('date_resolved_utc', models.DateTimeField(null=True, blank=True)),
                ('date_resplan_utc', models.DateTimeField(null=True, blank=True)),
                ('date_responded_utc', models.DateTimeField(null=True, blank=True)),
                ('is_in_sla', models.NullBooleanField()),
                ('api_text', models.TextField(null=True, blank=True)),
                ('board_name', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('board_id', models.IntegerField(null=True, blank=True)),
                ('board_status_id', models.IntegerField(null=True, blank=True)),
                ('company', models.ForeignKey(blank=True, null=True,
                                              related_name='company_tickets', to='djconnectwise.Company')),
            ],
            options={
                'verbose_name_plural': 'Service Tickets',
                'verbose_name': 'Service Ticket',
            },
        ),
        migrations.CreateModel(
            name='ServiceTicketAssignment',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        serialize=False, verbose_name='ID', primary_key=True)),
                ('created', django_extensions.db.fields.CreationDateTimeField(
                    auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(
                    auto_now=True, verbose_name='modified')),
                ('member', models.ForeignKey(to='djconnectwise.Member')),
                ('service_ticket', models.ForeignKey(
                    to='djconnectwise.ServiceTicket')),
            ],
            options={
                'get_latest_by': 'modified',
                'ordering': ('-modified', '-created'),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SyncJob',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        serialize=False, verbose_name='ID', primary_key=True)),
                ('start_time', models.DateTimeField(auto_now_add=True)),
                ('end_time', models.DateTimeField(null=True, blank=True)),
            ],
        ),
        migrations.CreateModel(
            name='TicketPriority',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        serialize=False, verbose_name='ID',
                                        primary_key=True)),
                ('created', django_extensions.db.fields.CreationDateTimeField(
                    auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(
                    auto_now=True, verbose_name='modified')),
                ('title', models.CharField(verbose_name='title', max_length=255)),
                ('description', models.TextField(null=True,
                                                 blank=True, verbose_name='description')),
            ],
            options={
                'get_latest_by': 'modified',
                'ordering': ('-modified', '-created'),
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TicketStatus',
            fields=[
                ('id', models.AutoField(auto_created=True,
                                        serialize=False, verbose_name='ID', primary_key=True)),
                ('created', django_extensions.db.fields.CreationDateTimeField(
                    auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(
                    auto_now=True, verbose_name='modified')),
                ('status_id', models.IntegerField(
                    null=True, blank=True, unique=True)),
                ('ticket_status', models.CharField(
                    null=True, blank=True, max_length=250)),
                ('status_name', models.CharField(
                    null=True, blank=True, max_length=250)),
            ],
            options={
                'get_latest_by': 'modified',
                'ordering': ('-modified', '-created'),
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='serviceticket',
            name='members',
            field=models.ManyToManyField(
                related_name='member_tickets', to='djconnectwise.Member', through='djconnectwise.ServiceTicketAssignment'),
        ),
        migrations.AddField(
            model_name='serviceticket',
            name='priority',
            field=models.ForeignKey(
                blank=True, null=True, to='djconnectwise.TicketPriority'),
        ),
        migrations.AddField(
            model_name='serviceticket',
            name='project',
            field=models.ForeignKey(
                blank=True, null=True, related_name='project_tickets', to='djconnectwise.Project'),
        ),
        migrations.AddField(
            model_name='serviceticket',
            name='status',
            field=models.ForeignKey(
                blank=True, null=True, related_name='status_tickets', to='djconnectwise.TicketStatus'),
        ),
        migrations.AddField(
            model_name='member',
            name='service_provider',
            field=models.ForeignKey(
                related_name='members', to='djconnectwise.ServiceProvider'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0007_auto_20170215_1918'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='company',
            options={'verbose_name_plural': 'companies', 'ordering': ('company_identifier',)},
        ),
        migrations.AlterModelOptions(
            name='member',
            options={'ordering': ('first_name', 'last_name')},
        ),
        migrations.AlterModelOptions(
            name='project',
            options={'ordering': ('name',)},
        ),
        migrations.AlterModelOptions(
            name='serviceticket',
            options={'verbose_name_plural': 'Service Tickets', 'verbose_name': 'Service Ticket', 'ordering': ('summary',)},
        ),
        migrations.AlterModelOptions(
            name='ticketpriority',
            options={'ordering': ('name',)},
        ),
        migrations.AlterModelOptions(
            name='ticketstatus',
            options={'verbose_name_plural': 'ticket statuses', 'ordering': ('ticket_status',)},
        ),
    ]



import os
import re
import urllib.error
import urllib.parse
import urllib.request

from easy_thumbnails.fields import ThumbnailerImageField
from model_utils import Choices

from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django_extensions.db.models import TimeStampedModel


PRIORITY_RE = re.compile('^Priority ([\d]+)')


class SyncJob(models.Model):
    start_time = models.DateTimeField(auto_now_add=True)
    end_time = models.DateTimeField(blank=True, null=True)


class CallBackEntry(models.Model):
    CALLBACK_TYPES = Choices(
        ('ticket', "Ticket"),
    )

    callback_type = models.CharField(max_length=25)
    url = models.CharField(max_length=255)
    level = models.CharField(max_length=255)
    object_id = models.IntegerField()
    entry_id = models.IntegerField()
    member_id = models.IntegerField()
    enabled = models.BooleanField(default=False)

    def __str__(self):
        return self.url


class ConnectWiseBoard(TimeStampedModel):
    board_id = models.PositiveSmallIntegerField()
    name = models.CharField(max_length=255)
    inactive = models.BooleanField()

    class Meta:
        ordering = ('name',)

    def __str__(self):
        return self.name

    @property
    def board_statuses(self):
        return ConnectWiseBoardStatus.objects.filter(board_id=self.board_id)


class ConnectWiseBoardStatus(TimeStampedModel):
    """
    Used for looking up the status/board id combination
    """
    board_id = models.PositiveSmallIntegerField()
    status_id = models.PositiveSmallIntegerField()
    status_name = models.CharField(blank=True, null=True, max_length=250)

    class Meta:
        ordering = ('status_name',)

    def __str__(self):
        return self.status_name


class Member(TimeStampedModel):
    identifier = models.CharField(
        max_length=15, blank=False, unique=True)  # This is the CW username
    first_name = models.CharField(max_length=30, blank=False)
    last_name = models.CharField(max_length=30, blank=False)
    office_email = models.EmailField(max_length=250)
    inactive = models.BooleanField(default=False)
    avatar = ThumbnailerImageField(null=True, blank=True, verbose_name=_(
        'Member Avatar'), help_text=_('Member Avatar'))

    class Meta:
        ordering = ('first_name', 'last_name')

    def __str__(self):
        return '{} {}'.format(self.first_name, self.last_name)

    def get_initials(self):
        name_segs = str(self).split(' ')
        initial = ''
        for seg in name_segs:
            seg = seg.strip()
            initial += seg[:1]

        return initial

    @staticmethod
    def create_member(api_member):
        member = Member()
        member.first_name = api_member['firstName']
        member.last_name = api_member['lastName']
        member.identifier = api_member['identifier']
        member.office_email = api_member['officeEmail']
        member.inactive = api_member['inactiveFlag']
        member.save()
        return member


class Company(TimeStampedModel):
    company_name = models.CharField(blank=True, null=True, max_length=250)
    company_alias = models.CharField(blank=True, null=True, max_length=250)
    company_identifier = models.CharField(
        blank=True, null=True, max_length=250)
    phone_number = models.CharField(blank=True, null=True, max_length=250)
    fax_number = models.CharField(blank=True, null=True, max_length=250)
    address_line1 = models.CharField(blank=True, null=True, max_length=250)
    address_line2 = models.CharField(blank=True, null=True, max_length=250)
    city = models.CharField(blank=True, null=True, max_length=250)
    state_identifier = models.CharField(blank=True, null=True, max_length=250)
    zip = models.CharField(blank=True, null=True, max_length=250)
    country = models.CharField(blank=True, null=True, max_length=250)
    type = models.CharField(blank=True, null=True, max_length=250)
    status = models.CharField(blank=True, null=True, max_length=250)
    territory = models.CharField(blank=True, null=True, max_length=250)
    website = models.CharField(blank=True, null=True, max_length=250)
    market = models.CharField(blank=True, null=True, max_length=250)
    defaultcontactid = models.IntegerField(blank=True, null=True)
    defaultbillingcontactid = models.IntegerField(blank=True, null=True)
    updatedby = models.CharField(blank=True, null=True, max_length=250)
    lastupdated = models.CharField(blank=True, null=True, max_length=250)

    class Meta:
        verbose_name_plural = 'companies'
        ordering = ('company_identifier', )

    def __str__(self):
        return self.get_company_identifier() or ''

    def get_company_identifier(self):
        identifier = self.company_identifier
        if settings.DJCONNECTWISE_COMPANY_ALIAS:
            identifier = self.company_alias or self.company_identifier
        return identifier


class TicketStatus(TimeStampedModel):
    CLOSED = 'Closed'

    status_id = models.IntegerField(blank=True, null=True, unique=True)
    # might ditch this field, it seems to always contain same value as
    # status_name
    ticket_status = models.CharField(blank=True, null=True, max_length=250)
    status_name = models.CharField(blank=True, null=True, max_length=250)

    class Meta:
        verbose_name_plural = 'ticket statuses'
        ordering = ('ticket_status',)

    def __str__(self):
        return self.status_name


class TicketPriority(TimeStampedModel):
    name = models.CharField(max_length=50, blank=False)
    priority_id = models.PositiveSmallIntegerField()
    # ConnectWise doesn't always return sort and color- not sure why.
    # Sort will be None in this circumstance- dependent code should handle it.
    sort = models.PositiveSmallIntegerField(null=True)
    # Color will be a property that tries to guess at a sensible value.
    _color = models.CharField(
        max_length=50, null=True, blank=True, db_column='color'
    )

    DEFAULT_COLORS = {
        '1': 'red',
        '2': 'orange',
        '3': 'yellow',
        '4': 'white',
        '5': 'darkmagenta',
    }
    DEFAULT_COLOR = 'darkgray'

    class Meta:
        verbose_name_plural = 'ticket priorities'
        ordering = ('name', )

    def __str__(self):
        return self.name

    @property
    def color(self):
        """
        If a color has been set, then return it. Otherwise if the name
        matches the common format ("Priority X - ..."), then return
        something sensible based on values seen in the wild.
        """
        if self._color:
            return self._color
        else:
            prio_number = None
            prio_match = PRIORITY_RE.match(self.name)
            if prio_match:
                prio_number = prio_match.group(1)
            return TicketPriority.DEFAULT_COLORS.get(
                prio_number,
                TicketPriority.DEFAULT_COLOR
            )

    @color.setter
    def color(self, color):
        self._color = color


class ServiceTicketAssignment(TimeStampedModel):
    service_ticket = models.ForeignKey('ServiceTicket')
    member = models.ForeignKey('Member')


class Project(TimeStampedModel):
    name = models.CharField(max_length=200)
    project_id = models.PositiveSmallIntegerField()
    project_href = models.CharField(max_length=200)

    class Meta:
        ordering = ('name', )

    def __str__(self):
        return self.name or ''


class ServiceTicket(TimeStampedModel):
    RECORD_TYPES = (
        ('ServiceTicket', "Service Ticket"),
        ('ProjectTicket', "Project Ticket"),
        ('ProjectIssue', "Project Issue"),
    )

    closed_flag = models.NullBooleanField(blank=True, null=True)
    type = models.CharField(blank=True, null=True, max_length=250)
    sub_type = models.CharField(blank=True, null=True, max_length=250)
    sub_type_item = models.CharField(blank=True, null=True, max_length=250)
    priority_text = models.CharField(blank=True, null=True, max_length=250)
    location = models.CharField(blank=True, null=True, max_length=250)
    source = models.CharField(blank=True, null=True, max_length=250)
    summary = models.CharField(blank=True, null=True, max_length=250)
    entered_date_utc = models.DateTimeField(blank=True, null=True)
    last_updated_utc = models.DateTimeField(blank=True, null=True)
    resources = models.CharField(blank=True, null=True, max_length=250)
    required_date_utc = models.DateTimeField(blank=True, null=True)
    closed_date_utc = models.DateTimeField(blank=True, null=True)
    site_name = models.CharField(blank=True, null=True, max_length=250)
    budget_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=6)
    actual_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=6)
    approved = models.NullBooleanField(blank=True, null=True)
    closed_by = models.CharField(blank=True, null=True, max_length=250)
    resolve_mins = models.IntegerField(blank=True, null=True)
    res_plan_mins = models.IntegerField(blank=True, null=True)
    respond_mins = models.IntegerField(blank=True, null=True)
    updated_by = models.CharField(blank=True, null=True, max_length=250)
    record_type = models.CharField(blank=True, null=True,
                                   max_length=250, choices=RECORD_TYPES,
                                   db_index=True)
    team_id = models.IntegerField(blank=True, null=True)
    agreement_id = models.IntegerField(blank=True, null=True)
    severity = models.CharField(blank=True, null=True, max_length=250)
    impact = models.CharField(blank=True, null=True, max_length=250)
    date_resolved_utc = models.DateTimeField(blank=True, null=True)
    date_resplan_utc = models.DateTimeField(blank=True, null=True)
    date_responded_utc = models.DateTimeField(blank=True, null=True)
    is_in_sla = models.NullBooleanField(blank=True, null=True)
    api_text = models.TextField(blank=True, null=True)
    board_name = models.CharField(blank=True, null=True, max_length=250)
    board_id = models.IntegerField(blank=True, null=True, db_index=True)
    board_status_id = models.IntegerField(blank=True, null=True)
    priority = models.ForeignKey('TicketPriority', blank=True, null=True)
    status = models.ForeignKey(
        'TicketStatus', blank=True, null=True, related_name='status_tickets')
    company = models.ForeignKey(
        'Company', blank=True, null=True, related_name='company_tickets')
    project = models.ForeignKey(
        'Project', blank=True, null=True, related_name='project_tickets')
    members = models.ManyToManyField(
        'Member', through='ServiceTicketAssignment',
        related_name='member_tickets')
    # TODO: add FK to ConnectWiseBoard

    class Meta:
        # ordering = ['priority_text','entered_date_utc','id']
        verbose_name = 'Service Ticket'
        verbose_name_plural = 'Service Tickets'
        ordering = ('summary', )

    def __str__(self):
        return '{}-{}'.format(self.id, self.summary)

    def get_connectwise_url(self):
        params = dict(
            locale='en_US',
            recordType='ServiceFv',
            recid=self.id,
            companyName=settings.CONNECTWISE_CREDENTIALS['company_id']
        )

        ticket_url = os.path.join(
            settings.CONNECTWISE_SERVER_URL,
            settings.CONNECTWISE_TICKET_PATH,
            '?{0}'.format(urllib.parse.urlencode(params))
        )

        return ticket_url

    def time_remaining(self):
        time_remaining = self.budget_hours
        if self.budget_hours and self.actual_hours:
            time_remaining = self.budget_hours - self.actual_hours
        return time_remaining



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0006_auto_20170130_1641'),
    ]

    operations = [
        migrations.AddField(
            model_name='ticketpriority',
            name='color',
            field=models.CharField(max_length=50, default='blue'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='ticketpriority',
            name='priority_id',
            field=models.PositiveSmallIntegerField(default=1),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='ticketpriority',
            name='sort',
            field=models.PositiveSmallIntegerField(default=0),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0005_auto_20170127_1549'),
    ]

    operations = [
        migrations.AlterField(
            model_name='serviceticket',
            name='board_id',
            field=models.IntegerField(null=True, blank=True, db_index=True),
        ),
        migrations.AlterField(
            model_name='serviceticket',
            name='record_type',
            field=models.CharField(max_length=250, null=True, blank=True, choices=[('ServiceTicket', 'Service Ticket'), ('ProjectTicket', 'Project Ticket'), ('ProjectIssue', 'Project Issue')], db_index=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djconnectwise', '0008_auto_20170215_1430'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='ticketpriority',
            options={'ordering': ('name',), 'verbose_name_plural': 'ticket priorities'},
        ),
    ]



# -*- coding: utf-8 -*-
from django.contrib import admin

from .models import Company, ConnectWiseBoard, Member
from .models import ServiceTicket, TicketStatus, TicketPriority


class ConnectWiseBoardAdmin(admin.ModelAdmin):
    model = ConnectWiseBoard
    list_display = ('name', 'board_id', 'inactive')
    search_fields = ['name']


class TicketStatusAdmin(admin.ModelAdmin):
    model = TicketStatus
    list_display = ('status_id', 'status_name')
    search_fields = ['status_name']


class MemberAdmin(admin.ModelAdmin):
    model = Member
    list_display = ('identifier', 'full_name', 'office_email')
    search_fields = ('identifier', 'first_name', 'last_name', 'office_email')

    def full_name(self, obj):
        return str(obj)


class CompanyAdmin(admin.ModelAdmin):
    model = Company
    list_display = ('id', 'company_name',
                    'company_identifier', 'type', 'status')
    list_filter = ('status',)
    search_fields = ['company_name', 'company_identifier']


class ServiceTicketAdmin(admin.ModelAdmin):
    model = ServiceTicket
    list_display = ('summary', 'status', 'resources', 'record_type',)

    list_filter = ('status', 'record_type',)
    search_fields = ['id', 'summary', 'members__user__username', ]

    def resources(self, obj):
        return ', '.join([str(m) for m in obj.members.all()])


class TicketPriorityAdmin(admin.ModelAdmin):
    model = TicketPriority
    list_display = ('priority_id', 'name', 'sort', 'color')


admin.site.register(ConnectWiseBoard, ConnectWiseBoardAdmin)
admin.site.register(TicketStatus, TicketStatusAdmin)
admin.site.register(Member, MemberAdmin)
admin.site.register(ServiceTicket, ServiceTicketAdmin)
admin.site.register(Company, CompanyAdmin)
admin.site.register(TicketPriority, TicketPriorityAdmin)
